'''
base_tab.py: base class for tabs

The tabs in Foliage are conceptually pretty simple: there's a function to
create the tab contents, and another function to set up optional watchers
for detecting and acting on designated PyWebIO "pin" objects.  The class
FoliageTab is a base class used by all the Foliage tab classes.  This common
base class makes it possible to implement tab creation and pin watching in
__main__.py's foliage_page() as a loop over a list of objects, rather than
by hardcoding calls to every tab directly.

Copyright
---------

Copyright (c) 2021-2022 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''


class FoliageTab():
    def contents(self):
        '''Return a dict of elements {'title': '...', 'content': [objects]}.'''
        raise NotImplementedError()

    def pin_watchers(self):
        '''Return a dict of elements {'pin_name': callback_function}.'''
        raise NotImplementedError()



'''
credentials.py: module for handling credentials for Foliage

Reading credentials
-------------------

This module implements storage and retrieval of credentials needed for the user
to interact with FOLIO.  It supports multiple ways of reading the creds:

- from environment variables
- from a .ini style file
- from a user interface dialog

Storing credentials
-------------------

The Foliage code only stores credentials outside of itself in one way: by
writing a combination of the FOLIO token, FOLIO tenant id, and FOLIO OKAPI URL
under the key "org.caltechlibrary.foliage" in the user's system keyring.
The values are never written to a file by Foliage.

Passing credentials around within Foliage
-----------------------------------------

The way credentials are communicated to the FOLIO functions (in folio.py) is
slightly unobvious.  The credentials are set via environment variables in the
Foliage process (see use_credentials() below) so that they're available to
the folio.py functions without having to pass around a variable holding them.
It's also a simpler alternative to keeping a singleton global object to hold
the credentials data.

When other functions below need to pass credentials around (e.g., when first
reading them in __main__.py), they're generally passed around in the form of a
named tuple, called (in a very imaginative fashion) "Credentials".  That's
done because while credentials are being configured or changed, the flow of
information is more clear when it's done point-to-point, via arguments.

Copyright
---------

Copyright (c) 2021-2022 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

from   collections import namedtuple
from   commonpy.interrupt import wait
from   decouple import Config, RepositoryIni, RepositoryEmpty, config
import getpass
import keyring
import os
from   pywebio.output import popup, close_popup, put_buttons, put_markdown
from   pywebio.output import put_loading
from   pywebio.pin import pin, put_input
from   sidetrack import log
import sys
import threading
from   validators.url import url as valid_url

if sys.platform.startswith('win'):
    import keyring.backends
    from keyring.backends.Windows import WinVaultKeyring
if sys.platform.startswith('darwin'):
    import keyring.backends
    from keyring.backends.OS_X import Keyring

from foliage.folio import Folio
from foliage.ui import confirm, note_info, notify


# Private constants.
# .............................................................................

_KEYRING = f'org.caltechlibrary.{__package__}'
'''The name of the keyring used to store server access credentials, if any.'''


# Public data types.
# .............................................................................

Credentials = namedtuple('Credentials', 'url tenant_id token')


# Public functions.
# .............................................................................

def credentials_from_file(creds_file):
    '''Return a Credentials object created from values in creds_file.

    The credentials file must be in ".ini" format, like this:

    [settings]
    FOLIO_OKAPI_URL = .....
    FOLIO_OKAPI_TENANT_ID = .....
    FOLIO_OKAPI_TOKEN = .....
    '''
    try:
        config_file = Config(RepositoryIni(source = creds_file))
    except Exception as ex:             # noqa: PIE786
        log('unable to read given creds file: ' + str(ex))
        return None
    return _creds_from_source(config_file, str(creds_file))


def credentials_from_env():
    '''Return a Credentials object created from environment variables.

    This checks for the following environment variables:

      FOLIO_OKAPI_URL
      FOLIO_OKAPI_TENANT_ID
      FOLIO_OKAPI_TOKEN
    '''
    return _creds_from_source(Config(RepositoryEmpty()), 'environment')


def credentials_from_user(warn_empty = True, initial_creds = None):
    '''Ask user for info needed to create a token & return a Credentials obj.'''

    event = threading.Event()
    clicked_ok = False

    def clk(val):
        nonlocal clicked_ok
        clicked_ok = val
        event.set()

    log('asking user for credentials')
    current = initial_creds or Credentials('', '', '')
    pins = [
        put_markdown('_This information is needed to create a FOLIO API token.'
                     ' Your FOLIO login & password will_'
                     ' **not** _be stored after this form disappears; only'
                     ' the token, URL and tenant id will be stored._'),
        put_input('user',      label = 'FOLIO user name'),
        put_input('password',  label = 'FOLIO password', type = 'password'),
        put_input('url',       label = 'OKAPI URL', value = current.url),
        put_input('tenant_id', label = 'Tenant id', value = current.tenant_id),
        put_buttons([
            {'label': 'Submit', 'value': True},
            {'label': 'Cancel', 'value': False, 'color': 'danger'},
        ], onclick = clk).style('float: right')
    ]
    popup(title = 'FOLIO credentials', content = pins, size = 'medium',
          closable = False)

    event.wait()
    close_popup()
    wait(0.5)                           # Give time for popup to go away.

    if not clicked_ok:
        log('user cancelled out of credentials dialog')
        return initial_creds

    if pin.url:                         # Remove '/' if the user included it.
        pin.url = pin.url.rstrip('/')

    if not all([pin.url, pin.tenant_id, pin.user, pin.password]):
        if warn_empty:
            log('user provided incomplete credentials')
            if confirm('Cannot proceed without all credentials. Try again?'):
                tmp = Credentials(url = pin.url, tenant_id = pin.tenant_id, token = None)
                return credentials_from_user(initial_creds = tmp)
            return None
        else:
            return None

    if not valid_url(pin.url):
        log('given URL that doesn\'t look like a uRL: ' + pin.url)
        if confirm('This does not look like a URL: "' + pin.url + '"'
                   + ' – would you like to edit the value and try again?'):
            tmp = Credentials(url = pin.url, tenant_id = pin.tenant_id, token = None)
            return credentials_from_user(initial_creds = tmp)
        else:
            return None

    with put_loading():
        token, error = Folio.new_token(url = pin.url, tenant_id = pin.tenant_id,
                                       user = pin.user, password = pin.password)

    if error:
        notify('Failed to get a token. ' + error + '.')
        return None
    else:
        note_info('New FOLIO API token obtained.')

    log('got credentials from user')
    return Credentials(url = pin.url, tenant_id = pin.tenant_id, token = token)


# Explanation about the weird way this is done: the Python keyring module
# only offers a single function for setting a value; ostensibly, this is
# intended to store a password associated with an identifier (a user name),
# and this identifier is expected to be obtained some other way, such as by
# using the current user's computer login name.  But, in our situation, we
# have multiple pieces of information we have to store (a user id and an api
# key).  The hackacious solution taken here is to concatenate the values into
# a single string used as the actual value stored.  The individual values are
# separated by a character that is unlikely to be part of any user-typed value.

def credentials_from_keyring(partial_ok = False, ring = _KEYRING):
    '''Look up the user's credentials.
    If partial_ok is False, return None if the keyring value is incomplete.'''
    if sys.platform.startswith('win'):
        log('using windows keyring vault')
        keyring.set_keyring(WinVaultKeyring())
    if sys.platform.startswith('darwin'):
        log('using macos keyring')
        keyring.set_keyring(Keyring())
    log(f'trying to read value from {ring}')
    try:
        value = keyring.get_password(ring, getpass.getuser())
    except Exception as ex:             # noqa: PIE786
        log('exception trying to get password from keyring: ' + str(ex))
        return None
    if value:
        log(f'got credentials from keyring {ring}')
        parts = _decoded(value)
        if all(parts) or partial_ok:
            return Credentials(url = parts[0], tenant_id = parts[1], token = parts[2])
    log(f'did not find a value in keyring {ring}')
    return None


def use_credentials(creds):
    '''Set run-time environment credentials and save them to the keyring.'''
    log('setting environment variables for credentials')
    os.environ['FOLIO_OKAPI_URL']       = creds.url
    os.environ['FOLIO_OKAPI_TENANT_ID'] = creds.tenant_id
    os.environ['FOLIO_OKAPI_TOKEN']     = creds.token
    if config('USE_KEYRING', cast = bool):
        keyring_creds = credentials_from_keyring()
        if creds != keyring_creds:
            _store_credentials(creds)


def current_credentials():
    url       = config('FOLIO_OKAPI_URL', default = None)
    tenant_id = config('FOLIO_OKAPI_TENANT_ID', default = None)
    token     = config('FOLIO_OKAPI_TOKEN', default = None)
    return Credentials(url = url, tenant_id = tenant_id, token = token)


def credentials_complete(creds):
    '''Return True if the given credentials are complete.'''
    return (creds and creds.url and creds.tenant_id and creds.token)


# Private helper functions.
# .............................................................................

_SEP = ''
'''Character used to separate multiple actual values stored as a single
encoded value string.  This character is deliberately chosen to be something
very unlikely to be part of a legitimate string value typed by user at a
shell prompt, because control-c is normally used to interrupt programs.
'''


def _encoded(url, tenant_id, token):
    return f'{url}{_SEP}{tenant_id}{_SEP}{token}'


def _decoded(value_string):
    return tuple(value_string.split(_SEP))


def _creds_from_source(source = None, where = ''):
    if not source:
        return None
    url       = source.get('FOLIO_OKAPI_URL', default = None)
    tenant_id = source.get('FOLIO_OKAPI_TENANT_ID', default = None)
    token     = source.get('FOLIO_OKAPI_TOKEN', default = None)
    if not any([url, tenant_id, token]):
        log(f'no credentials found in {where}')
        return None
    creds = Credentials(url = url, tenant_id = tenant_id, token = token)
    complete = 'complete' if credentials_complete(creds) else 'not complete'
    log(f'credentials in {where} are {complete}')
    return creds


def _store_credentials(creds, ring = _KEYRING):
    '''Save the user's credentials.'''
    if sys.platform.startswith('win'):
        keyring.set_keyring(WinVaultKeyring())
    if sys.platform.startswith('darwin'):
        keyring.set_keyring(Keyring())
    value = _encoded(creds.url, creds.tenant_id, creds.token)
    log(f'storing credentials to keyring {_KEYRING}')
    keyring.set_password(ring, getpass.getuser(), value)



def test_version():
    """Test version import."""
    from foliage import __version__
    assert __version__


def test_print_version(capsys):
    from foliage import print_version
    print_version()
    captured = capsys.readouterr()
    assert 'URL' in captured.out



'''
delete_tab.py: implementation of the "Delete records" tab

Copyright
---------

Copyright (c) 2021-2022 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

from   commonpy.exceptions import Interrupted
from   commonpy.interrupt import reset_interrupts, interrupt
from   decouple import config
import json
from   pywebio.output import put_markdown, put_row, put_button, use_scope
from   pywebio.output import put_grid, clear
from   pywebio.output import put_processbar, set_processbar
from   pywebio.output import put_scope, clear_scope
from   pywebio.pin import pin, put_textarea
from   sidetrack import log

from   foliage.base_tab import FoliageTab
from   foliage.exceptions import FolioOpFailed
from   foliage.export import export_data
from   foliage.folio import Folio, RecordKind, IdKind, TypeKind, Record
from   foliage.folio import unique_identifiers, back_up_record
from   foliage.ui import confirm, user_file, stop_processbar
from   foliage.ui import tell_success, tell_warning, tell_failure
from   foliage.ui import note_error, PROGRESS_BOX, PROGRESS_TEXT


# Tab definition class.
# .............................................................................

class DeleteTab(FoliageTab):
    def contents(self):
        return {'title': 'Delete records', 'content': tab_contents()}

    def pin_watchers(self):
        return {}


# Tab creation function.
# .............................................................................

def tab_contents():
    log('generating delete tab contents')
    return [
        put_grid([[
            put_markdown('Input one or more item, holdings, or instance'
                         ' identifiers (in the form of id\'s, barcodes, hrid\'s,'
                         ' or accession numbers), or upload a file containing'
                         ' the identifiers. **Warning**: Deleting holdings'
                         ' **will delete all their items**, and deleting'
                         ' instances **will delete all their holdings _and_ items**.'),
            put_button('Upload', outline = True,
                       onclick = lambda: load_file()).style('text-align: right'),
        ]], cell_widths = 'auto 100px'),
        put_textarea('textbox_delete', rows = 4),
        put_row([
            put_button('Delete records', onclick = lambda: do_delete()),
            put_button('Clear', outline = True,
                       onclick = lambda: clear_tab()).style('text-align: right')
        ])
    ]


# Implementation of tab functionality.
# .............................................................................

def clear_tab():
    log('clearing tab')
    clear('output')
    pin.textbox_delete = ''


def load_file():
    log('user requesting file upload')
    if (file := user_file('Upload a file containing identifiers')):
        pin.textbox_delete = file


def stop():
    log('stopping')
    interrupt()
    stop_processbar()


_results = []


def clear_results():
    global _results
    _results = []


def record_result(record_or_id, success, notes):
    global _results
    id_ = record_or_id if isinstance(record_or_id, str) else record_or_id.id
    rec = record_or_id if isinstance(record_or_id, Record) else None
    _results.append({'id': id_, 'success': success, 'notes': notes, 'record': rec})


def succeeded(record_or_id, msg, why = ''):
    comment = (' (' + why + ')') if why else ''
    record_result(record_or_id, True, msg + comment)
    tell_success('Success: ' + msg + comment + '.')


def failed(record_or_id, msg, why = ''):
    comment = (' (' + why + ')') if why else ''
    record_result(record_or_id, False, msg + comment)
    id_ = record_or_id if isinstance(record_or_id, str) else record_or_id.id
    tell_failure(f'Failed to delete **{id_}**{comment}: ' + msg + '.')


def skipped(record_or_id, msg, why = ''):
    comment = (' (' + why + ')') if why else ''
    record_result(record_or_id, False, msg + comment)
    id_ = record_or_id if isinstance(record_or_id, str) else record_or_id.id
    tell_warning(f'Skipped **{id_}**{comment}: ' + msg + '.')


def flagged(record_or_id, msg, why = ''):
    comment = (' (' + why + ')') if why else ''
    id_ = record_or_id if isinstance(record_or_id, str) else record_or_id.id
    tell_warning(f'Note about **{id_}**{comment}: ' + msg + '.')


def do_delete():
    log('do_delete invoked')
    identifiers = unique_identifiers(pin.textbox_delete)
    if not identifiers:
        note_error('Please input at least one barcode or other type of id.')
        return
    if not confirm('Warning: if the deletions include holdings and/or instance'
                   ' records, all their associated items and holdings will'
                   ' also be deleted. Only do this if you have verified the'
                   ' implications first. Proceed?', danger = True):
        log('user declined to proceed')
        return
    clear_results()
    reset_interrupts()
    steps = 2*len(identifiers)       # Count getting records, for more action.
    folio = Folio()
    with use_scope('output', clear = True):
        put_grid([[
            put_scope('current_activity', [
                put_markdown('_Getting records ..._').style(PROGRESS_TEXT)]),
        ], [
            put_processbar('bar', init = 0/steps).style('margin-top: 11px'),
            put_button('Stop', outline = True, color = 'danger',
                       onclick = lambda: stop()).style('text-align: right')
        ]], cell_widths = '85% 15%').style(PROGRESS_BOX)
        try:
            done = 0
            for id_ in identifiers:
                with use_scope('current_activity', clear = True):
                    put_markdown(f'_Looking up {id_} ..._').style(PROGRESS_TEXT)
                record = folio.record(id_)
                if not record:
                    failed(id_, f'unrecognized identifier **{id_}**')
                    continue
                done += 1
                set_processbar('bar', done/steps)
                if record.kind not in _HANDLERS.keys():
                    skipped(id_, f'deleting {record.kind} records is not supported')
                    continue
                with use_scope('current_activity', clear = True):
                    text = '_Deleting '
                    if record.kind is RecordKind.ITEM:
                        text += f'item {id_} ..._'
                    elif record.kind is RecordKind.HOLDINGS:
                        text += f'holdings {id_} and associated items ..._'
                    elif record.kind is RecordKind.INSTANCE:
                        text += f'instance {id_} and associated holdings and items ..._'
                    elif record.kind is RecordKind.USER:
                        text += f'user {id_} and associated holdings and items ..._'
                    put_markdown(text).style(PROGRESS_TEXT)
                _HANDLERS[record.kind](record)
                done += 1
                set_processbar('bar', done/steps)
            set_processbar('bar', 1)
            clear_scope('current_activity')
        except Interrupted:
            tell_warning('**Stopped**.')
            return
        except Exception as ex:         # noqa: PIE786
            import traceback
            log('Exception: ' + str(ex) + '\n' + traceback.format_exc())
            tell_failure('Error: ' + str(ex))
            return
        finally:
            stop_processbar()

        put_grid([[
            put_markdown('Finished deletions.').style('margin-top: 6px'),
            put_button('Export summary', outline = True,
                       onclick = lambda: do_export('foliage-deletions.csv'),
                       ).style('text-align: right')
        ]]).style('margin: 1.5em 17px auto 17px')


def delete(record, for_id = None):
    '''Generic low-level function to delete the given record.'''
    why = ('for request to delete ' + for_id) if for_id else ''
    if config('DEMO_MODE', cast = bool):
        log(f'demo mode in effect – pretending to delete {record.id}')
    else:
        try:
            back_up_record(record)
            folio = Folio()
            folio.delete_record(record)
        except FolioOpFailed as ex:
            failed(record, str(ex), why)
            return False
    succeeded(record, f'deleted {record.kind} record **{record.id}**', why)
    return True


def delete_holdings(holdings, for_id = None):
    '''Delete the given holdings record and all its items.'''
    # Deletions on holdings are not recursive. The items have to be deleted 1st
    # and FOLIO will give a 400 error if you try to delete a holdings record
    # while there's still an item somewhere pointing to it.
    folio = Folio()
    # Start at the bottom: delete its items 1st.
    for item in folio.related_records(holdings.id, IdKind.HOLDINGS_ID, RecordKind.ITEM):
        if not delete(item, for_id = holdings.id):
            failed(holdings, f'unable to delete item {item.id} – stopping')
            return False
    # If we get this far, delete the holdings record.
    return delete(holdings, for_id)


def delete_instance(instance, for_id = None):
    '''Delete the given instance record.'''
    why = f'for request to delete FOLIO instance record {instance.id}'
    folio = Folio()

    # Deletions must be done in two separate places: source storage (SRS), and
    # the regular instance/holdings/items storage APIs. Kyle Banerjee said in
    # personal communication of 2022-07-27 that SRS should be done first.

    def srs_response_converter(response):
        if not response or not response.text:
            log('no response received from FOLIO')
            return None
        if response.status_code == 404:
            # Assume this is a case of an instance lacking a Marc record.
            log('No SRS record found')
            return None
        return json.loads(response.text)

    srsget = f'/source-storage/records/{instance.id}/formatted?idType=INSTANCE'
    data_json = folio.request(srsget, converter = srs_response_converter)
    if not data_json:
        flagged(instance, ("FOLIO SRS lacks a corresponding record, therefore"
                           " only the instance record will be deleted"))
    elif 'matchedId' not in data_json:
        failed(instance, 'unexpected data from FOLIO SRS – please report this')
        return
    else:
        srs_id = data_json["matchedId"]
        if config('DEMO_MODE', cast = bool):
            log(f'demo mode in effect – pretending to delete {srs_id} from SRS')
        else:
            try:
                log(f'deleting {instance.id} from SRS, where its id is {srs_id}')
                srsdel = f'/source-storage/records/{srs_id}'
                folio.request(srsdel, op = 'delete')
            except FolioOpFailed as ex:
                failed(instance, str(ex), why)
                return False
        succeeded(instance, f'removed SRS instance record **{srs_id}**', why)

    # Deletions on instances are not recursive regardless of which API you use.
    # You have to manually remove items, then holdings, then instances.
    # The following is based on Kyle Banerjee's script dated 2021-11-11 at
    # https://github.com/FOLIO-FSE/shell-utilities/blob/master/instance-delete.

    # Start by using delete_holdings(), which will delete items too.
    folio = Folio()
    for holdings in folio.related_records(instance.id, IdKind.INSTANCE_ID,
                                          RecordKind.HOLDINGS):
        if not delete_holdings(holdings, for_id = instance.id):
            failed(instance, 'unable to delete all holdings records')
            return False

    # If we didn't get an exception, finally delete the instance from inventory.
    return delete(instance, for_id)


def delete_user(record, for_id = None):
    '''Delete the given user record.'''
    failed(record, 'user records cannot be deleted at this time')


_location_map = None


def init_location_map():
    global _location_map
    if _location_map is None:
        folio = Folio()
        _location_map = {x.data['id']: x.data['name']
                         for x in folio.types(TypeKind.LOCATION)}


def location(location_data):
    global _location_map
    if isinstance(location_data, dict):
        if 'name' in location_data:
            return f'{location_data["name"]} ({location_data["id"]})'
        else:
            return location_data["id"]
    elif location_data and location_data in _location_map:
        return f'{_location_map[location_data]} ({location_data})'
    return '(unknown location)'


def do_export(file_name):
    global _results
    init_location_map()
    # Output fields requested
    #   id
    #   success
    #   notes
    #   record hrid
    #   record barcode (if present)
    #   record location
    #   record title
    #   record publication year
    #   record publisher

    # People want to see titles, dates, locations. We need to do lookups to
    # get them.
    values = []
    for result in _results:
        entry = {'Record ID'          : result['id'],
                 'Operation success'  : result['success'],
                 'Notes'              : result['notes'],
                 'Effective location' : '',
                 'Permanent location' : ''}
        if result['record']:
            rec = result['record']
            if rec.kind == RecordKind.ITEM:
                entry['Permanent location'] = location(rec.data['permanentLocationId'])
                if 'effectiveLocationId' in rec.data:
                    entry['Effective location'] = location(rec.data['effectiveLocationId'])
        values.append(entry)
    export_data(values, file_name, sort = False)


_HANDLERS = {
    RecordKind.ITEM     : delete,       # The generic function suffices.
    RecordKind.HOLDINGS : delete_holdings,
    RecordKind.INSTANCE : delete_instance,
    RecordKind.USER     : delete_user,
}



'''
export.py: let the user export records and save them to a file

Copyright
---------

Copyright (c) 2021-2022 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

from   commonpy.data_utils import pluralized, flattened
from   commonpy.interrupt import wait
import csv
from   io import BytesIO, StringIO
import json
from   pywebio.output import popup, close_popup, put_buttons
from   pywebio.pin import pin, put_radio
from   pywebio.session import download
from   sidetrack import log
from   slugify import slugify
import threading

from   foliage.folio import RecordKind
from   foliage.ui import note_error, note_warn


# Main functions.
# .............................................................................

def export_records(records, kind):
    log(f'exporting {pluralized(kind + " record", records, True)}')
    if not records:
        note_error('No results – nothing to export.')
        return

    event = threading.Event()
    clicked_ok = False

    def clk(val):
        nonlocal clicked_ok
        clicked_ok = val
        event.set()

    log('asking user for output format')
    pins = [
        put_radio('file_fmt', options = [('CSV', 'csv', True), ('JSON', 'json')]),
        put_buttons([
            {'label': 'Cancel', 'value': False, 'color': 'secondary'},
            {'label': 'OK', 'value': True},
        ], onclick = clk).style('float: right; vertical-align: center')
    ]
    popup(title = 'Select the format for the exported records:', content = pins)

    event.wait()
    close_popup()
    wait(0.5)                           # Give time for popup animation.

    if not clicked_ok:
        log('user clicked cancel')
        return

    if pin.file_fmt == 'csv':
        log('user selected CSV format')
        export_records_csv(records, kind)
    else:
        log('user selected JSON format')
        export_records_json(records, kind)


def export_data(data_list, filename, sort = True):
    if not data_list:
        return

    # Assume all the items have the same structure and it is flat.
    columns = list(data_list[0].keys())
    if sort:
        # Try to find a good sort key, else default to 1st key found.
        if 'id' in columns:
            sort_key = 'id'
        elif 'record id' in columns:
            sort_key = 'Record ID'
        elif 'name' in columns:
            sort_key = 'name'
        else:
            sort_key = list(data_list[0].keys())[0]
        data_list = sorted(data_list, key = lambda d: d[sort_key])

    # Write into an in-memory, file-like object & tell PyWebIO to download it.
    with StringIO() as tmp:
        writer = csv.DictWriter(tmp, fieldnames = columns)
        writer.writeheader()
        for item_dict in data_list:
            writer.writerow(item_dict)
        tmp.seek(0)
        bytes_ = BytesIO(tmp.read().encode('utf8')).getvalue()
        download(filename, bytes_)


# Miscellaneous helper functions.
# .............................................................................

def export_records_csv(records, kind):
    log(f'exporting {pluralized("record", records, True)} to CSV')
    if len(records) == 0:
        note_warn('List of records is empty.')

    # We have nested dictionaries, which can't be stored directly in CSV, so
    # first we have to flatten the dictionaries inside the list.
    records = [flattened(r.data) for r in records]

    # Next, we need a list of column names to pass to the CSV function.  This
    # is complicated by the fact that JSON dictionaries can have fields that
    # themselves have JSON dictionaries for values, and any given record (1)
    # may not have values for all those fields, and (2) may have values that
    # are lists, but with different numbers of elements. So we can't just
    # look at one record to figure out all the columns we need: we have to
    # look at _all_ records and create a maximal set before we write the CSV.
    columns = set(flattened(record.keys() for record in records))

    # Sort the column names to move the name & id fields to the front.
    name_key = RecordKind.name_key(kind)
    # FIXME storage records don't have the same fields as inventory records.
    # RecordKind.name_key needs a way to make that distinction.
    if name_key not in records[0]:
        name_key = 'id'

    def name_id_key(column_name):
        return (column_name != name_key, column_name != 'id', column_name)
    columns = sorted(list(columns), key = lambda x: name_id_key(x))

    # Write into an in-memory, file-like object & tell PyWebIO to download it.
    with StringIO() as tmp:
        writer = csv.DictWriter(tmp, fieldnames = columns)
        writer.writeheader()
        for item_dict in sorted(records, key = lambda d: d[name_key]):
            writer.writerow(item_dict)
        tmp.seek(0)
        bytes_ = BytesIO(tmp.read().encode('utf8')).getvalue()
        download(f'{slugify(kind)}-records.csv', bytes_)


def export_records_json(records, kind):
    log(f'exporting {pluralized("record", records, True)} to JSON')
    if len(records) == 0:
        note_warn('List of records is empty.')
    records_json = [r.data for r in records]
    with StringIO() as tmp:
        json.dump(records_json, tmp)
        tmp.seek(0)
        bytes_ = BytesIO(tmp.read().encode('utf8')).getvalue()
        download(f'{slugify(kind)}-records.json', bytes_)



'''
__init__.py for foliage

Copyright
---------

Copyright (c) 2021-2022 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

# Package metadata
# .............................................................................
#
#  ╭────────────────────── Notice ── Notice ── Notice ─────────────────────╮
#  |    The following values are automatically updated at every release    |
#  |    by the Makefile. Manual changes to these values will be lost.      |
#  ╰────────────────────── Notice ── Notice ── Notice ─────────────────────╯

__version__     = '1.6.0'
__description__ = 'Foliage: a tool to do bulk changes in FOLIO using the OKAPI API'
__url__         = 'https://github.com/caltechlibrary/foliage'
__author__      = 'Mike Hucka'
__email__       = 'helpdesk@library.caltech.edu'
__license__     = 'BSD 3-clause license'


# Miscellaneous utilities.
# .............................................................................

def print_version():
    print(f'{__name__} version {__version__}')
    print(f'Authors: {__author__}')
    print(f'URL: {__url__}')
    print(f'License: {__license__}')



def test_encoding():
    from foliage.credentials import _encoded, _decoded
    assert _decoded(_encoded('a', '1', 'c')) == ('a', '1', 'c')


def test_current_credentials(monkeypatch):
    from foliage.credentials import Credentials, current_credentials
    monkeypatch.setenv('FOLIO_OKAPI_URL', 'https://foo')
    monkeypatch.setenv('FOLIO_OKAPI_TENANT_ID', '1')
    monkeypatch.setenv('FOLIO_OKAPI_TOKEN', 'abc')
    assert current_credentials() == Credentials('https://foo', '1', 'abc')


def test_credentials_complete():
    from foliage.credentials import Credentials, credentials_complete
    assert credentials_complete(Credentials('https://foo', '1', 'abc'))


def test_credentials_from_file(monkeypatch, tmp_path):
    from foliage.credentials import Credentials, credentials_from_file
    d = tmp_path / 'sub'
    d.mkdir()
    p = d / 'settings.ini'
    p.write_text('''
[settings]
FOLIO_OKAPI_URL = https://foo
FOLIO_OKAPI_TENANT_ID = 1
FOLIO_OKAPI_TOKEN = abc
''')
    monkeypatch.chdir(d)
    assert credentials_from_file(str(p)) == Credentials('https://foo', '1', 'abc')


def test_credentials_from_env(monkeypatch):
    from foliage.credentials import Credentials, credentials_from_env
    monkeypatch.setenv('FOLIO_OKAPI_URL', 'https://foo')
    monkeypatch.setenv('FOLIO_OKAPI_TENANT_ID', '1')
    monkeypatch.setenv('FOLIO_OKAPI_TOKEN', 'abc')
    assert credentials_from_env() == Credentials('https://foo', '1', 'abc')



'''
__main__.py: main function for foliage.

User interface design
---------------------

Foliage runs locally as a Python application, but it opens a web page in the
user's default web browser to use as a user interface.  Foliage does this via
PyWebIO, a framework for building browser-based synchronous GUI applications.
The page has familiar GUI elements like page tabs, buttons, and input forms.

The main Foliage application also starts a separate graphical interface written
in PyQT to put a widget in the taskbar (on Windows) or system tray (on macOS),
This widget provides a visual indication to the user that Foliage is still
running, and also provides a single menu item, "Quit", that allows the user to
quit Foliage.  It's an alternative to using the "Quit" button in the main
Foliage window or closing the window, in case the user has done something to
hide or lose track of the Window.  On Windows, the PyQT widget runs in a thread
but on macOS, it must run as a separate process spawned by Foliage.  The macOS
requirement is a result of how macOS's graphical framework works; the issues
are explained in the file "system_widget.py" in this directory.

The PyWebIO framework does its magic by running a web server on the user's
machine.  Foliage uses Tornado as the web server, mainly because Tornado
supports a useful feature during development: auto-reloading Foliage files
when they've been edited.  However, the PyWebIO interface to Tornado lacks a
critical feature needed by Foliage to implement a proper "Quit" capability
(which is a significant problem, because being able to quit the app is kind
of an important thing). The copy of PyWebIO currently used by Foliage is a
modified version PyWebIO 1.5.2; the modification solves a problem I was having
in which PyWebIO could not detect that the user has closed the application
window.  I'm submitting a pull request to PyWebIO that I hope will mean future
versions of PyWebIO will provide the feature and Foliage won't have to use a
fork in the future.

The main event loop for user interaction is implemented in the Foliage function
foliage_page() in this file.  This function populates the web page (calling
on functions in other files to create the elements on the various tabs) and
then ends with a "while True" loop.  The loop periodically checks whether
certain GUI elements have changed state (such as certain buttons being pressed)
and if not, goes back to waiting for a change in state.  Most of the tabs in
Foliage actually don't need to be engaged in this loop, because they set up
buttons to trigger functions if the user presses the button.  However, not
everything can be done that way, plus we need to check the state of the widget
mentioned above and that needs to be done by polling.

FOLIO user credentials
----------------------

The typical scenario expected for Foliage is that the first time the user
starts it, it will ask the user for credentials, store a token in the user's
private system keychain, and then on subsequent starts, Foliage will read the
token the keychain instead of asking the user again.  The information Foliage
needs is a FOLIO tenant id, FOLIO URL, and (once only to create a token) the
user's FOLIO login and password; it does not store the user login and password
because it only needs the token, tenant id and URL.

For flexibility and to support other scenarios, Foliage actually supports
multiple ways of providing the credentials:

  1. If the command line argument --creds-file is used, and the given file
     contains the credentials (OKAPI URL, tenant id, and OKAPI API token), then
     those credentials are used and nothing more is tried. (In other words,
     credentials given via --creds-file override everything else.) If the
     given file cannot be read or does not contain complete credentials, then
     it is an error and Foliage will quit.

  2. Else, if all three of the environment variables FOLIO_OKAPI_URL,
     FOLIO_OKAPI_TENANT_ID and FOLIO_OKAPI_TOKEN are set, it uses those
     credentials and nothing more is tried.

  3. Else, it looks up the credentials in the user's system keyring/keychain.
     If all 3 pieces of info are found (i.e., OKAPI URL, tenant id, and OKAPI
     API token), then those credentials are used.

  4. Else, it asks the user for the FOLIO_OKAPI_URL, FOLIO_OKAPI_TENANT_ID, a
     FOLIO login, and a password, and uses that combination to request an API
     token from FOLIO. Unless the --no-keyring argument is given, Foliage also
     stores the URL, tenant id, and token in the user's keyring/keychain so
     that it doesn't have to ask again in future runs. (The user's login and
     password are not stored.)

FOLIO API
---------

Calls to the FOLIO API are encapsulated in folio.py and credentials.py.
This main function doesn't interact directly with FOLIO except during the
initial startup when it deals with user credentials.  The calls to FOLIO are
mostly all made by the various tab functions.

Debug log
---------

Unlike some other applications I've written, Foliage always writes a log file.
The file can be accessed by the user from the "Other" tab.  The location of
the file is set to a default application-specific location unless the user
sets it explicitly to another destination using the command-line argument
--debug.  Using the command-line argument is the only way to set it; the GUI
does not provide a way to change the location.  (This is done on purpose;
changing the log destination is something only advanced users or developers
should do.  For normal users, we want the output to go to a known and reliable
location, so that it's findable in case it's needed to help solve problems
that users may be experiencing.)

Backup directory
----------------

Before destructive operations are made on records in FOLIO, Foliage saves a
copy of the record as it existed before the change is performed.  It saves it
in an application-specific directory.  Similar to the case of the debug log
destination, the backup directory can also be changed via a command-line
argument, --backup-dir.  For similar reasons to those discussed above for the
debug flag, there is no GUI facility to change the location of the backup dir.
It can only be done via the command line interface.

Copyright
---------

Copyright (c) 2021-2022 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

import sys
if sys.version_info <= (3, 8):
    print('foliage requires Python version 3.8 or higher,')
    print('but the current version of Python is '
          + str(sys.version_info.major) + '.' + str(sys.version_info.minor) + '.')
    sys.exit(1)

from   appdirs import AppDirs
from   collections import ChainMap
from   commonpy.data_utils import timestamp
from   commonpy.exceptions import Interrupted
from   commonpy.file_utils import writable, readable
from   commonpy.interrupt import config_interrupt, wait
from   commonpy.network_utils import network_available
from   commonpy.string_utils import antiformat
from   decouple import config
from   fastnumbers import isint
import faulthandler
from   functools import partial
import os
from   os import makedirs
from   os.path import exists, dirname, join, realpath, isdir
import plac
import pywebio
# PyWebIO's default is tornado, and tornado mucks with logging, so I would
# have preferred to use aiohttp (which doesn't muck with logging).
# Unfortunately, only tornado provides an auto-reload feature.  IMPORTANT:
# the copy of PyWebIO used by Foliage is a fork where I've modified an
# important function for detecting when the user has closed the app window.
from   pywebio import start_server
from   pywebio.output import put_html, put_warning, put_tabs, put_image
from   pywebio.pin import pin_wait_change, put_actions
from   pywebio.session import run_js
from   sidetrack import set_debug, log
from   tornado.template import Template

from   foliage import __version__
from   foliage.change_tab import ChangeTab
from   foliage.credentials import credentials_from_user, credentials_from_keyring
from   foliage.credentials import use_credentials, credentials_complete
from   foliage.credentials import credentials_from_file, credentials_from_env
from   foliage.delete_tab import DeleteTab
from   foliage.folio import Folio
from   foliage.list_tab import ListTab
from   foliage.lookup_tab import LookupTab
from   foliage.other_tab import OtherTab
from   foliage.clean_tab import CleanTab
from   foliage.system_widget import SystemWidget
from   foliage.ui import quit_app, confirm, notify, inside_pyinstaller_app
from   foliage.ui import note_info, note_warn, note_error
from   foliage.ui import image_data, JS_CODE, CSS_CODE
from   foliage.ui import close_splash_screen


# Internal constants.
# .............................................................................

_DIRS = AppDirs('Foliage', 'CaltechLibrary')
'''Platform-specific directories for Foliage data.'''

_TABS = [LookupTab(), ChangeTab(), DeleteTab(), CleanTab(), ListTab(), OtherTab()]
'''List of tabs making up the Foliage application.'''


# Main program.
# .............................................................................

@plac.annotations(
    backup_dir = ('back up records to folder B before changes', 'option', 'b'),
    creds_file = ('read FOLIO credentials from .ini file C'   , 'option', 'c'),
    demo_mode  = ('demo mode: don\'t perform destructive ops' , 'flag'  , 'd'),
    no_keyring = ('don\'t use keyring for credentials'        , 'flag'  , 'K'),
    port       = ('open browser on port P (default: 8080)'    , 'option', 'p'),
    version    = ('print program version info and exit'       , 'flag'  , 'V'),
    no_widget  = ('don\'t run the taskbar/system tray widget' , 'flag'  , 'W'),
    debug      = ('log debug output to "OUT" ("-" is console)', 'option', '@'),
)
def main(backup_dir = 'B', creds_file = 'C', demo_mode = False,
         no_keyring = False, port = 'P', version = False, no_widget = False,
         debug = 'OUT'):
    '''
Foliage (FOLIo chAnGe Editor) is a tool to do bulk changes in FOLIO. It allows
a user to look up records of various kinds, perform bulk changes in the values
of record fields, delete records, and more. It communicates with a FOLIO server
using the OKAPI network API.

FOLIO credentials
~~~~~~~~~~~~~~~~~

Credentials for FOLIO OKAPI access can be provided in a number of ways. The
sequence of methods it uses are as follows:

  1. If the --creds-file argument is used, and the given file contains
     complete credentials (OKAPI URL, tenant id, and OKAPI API token), then
     those credentials are used and nothing more is tried. (In other words,
     credentials given via --creds-file override everything else.) If the
     given file cannot be read or does not contain complete credentials, then
     it is an error and Foliage will quit.

  2. Else, if all three of the environment variables FOLIO_OKAPI_URL,
     FOLIO_OKAPI_TENANT_ID and FOLIO_OKAPI_TOKEN are set, it uses those
     credentials and nothing more is tried.

  3. Else, the credentials are looked up in the user's system keyring/keychain.
     If all 3 pieces of info are found (i.e., OKAPI URL, tenant id, and OKAPI
     API token), then those credentials are used.

  4. Else, it asks the user for the FOLIO_OKAPI_URL, FOLIO_OKAPI_TENANT_ID, a
     FOLIO login, and a password, and uses that combination to request an API
     token from FOLIO. Unless the --no-keyring argument is given, Foliage also
     stores the URL, tenant id, and token in the user's keyring/keychain so
     that it doesn't have to ask again in future runs. (The user's login and
     password are not stored.)

In normal situations, the first time a user runs Foliage, they will end up in
case #4 above (i.e., being asked for credentials so that Foliage can create
and store an API token), and then on subsequent runs, Foliage will not ask for
the credentials again.

The form of the interface
~~~~~~~~~~~~~~~~~~~~~~~~~

Although Foliage is a desktop application and not a web service, it uses a
web page as its user interface -- it opens a page in a browser on your
computer, letting you interact with the program through the familiar elements
of a web page. All the while, Foliage runs locally on your computer. When you
start Foliage normally (or after it shows the one-time credentials screen,
described below), your browser should present a page that has the title
"Foliage" and is organized into five areas of functionality accessed by
clicking on the row of tabs near the top: (1) "Look up records" (the first
one shown when Foliage starts up), (2) "Change records", (3) "Delete
records", (4) "List UUIDs", and (5) "Other".

Taskbar/menubar widget
~~~~~~~~~~~~~~~~~~~~~~

When Foliage starts up, it provides an icon in the Windows taskbar or the
macOS system tray (depending on your operating system). The icon serves as a
reminder that Foliage is running, and offers a single menu option (for
quitting Foliage). The menu is accssed by right-clicking the widget on
Windows or left-clicking it on macOS.

To prevent the startup of the widget process, run Foliage with the --no-widget
command-line option.

Additional command-line arguments
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

The default port for the local web server started by Foliage is 8080.  To
change this port, you can use the option --port followed by a port number.

If given the -V option, this program will print the version and other
information, and exit without doing anything else.

If given the -@ argument, this program will output a detailed trace of what it
is doing. The debug trace will be sent to the given destination, which can
be '-' to indicate console output, or a file path to send the output to a file.

Command-line arguments summary
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'''

    # Process arguments -------------------------------------------------------

    if version:
        from foliage import print_version
        print_version()
        sys.exit()

    config_debug(debug)                # Set up debugging before going further.

    log('='*8 + f' started {timestamp()} ' + '='*8)
    log('command line: ' + str(sys.argv))
    os.environ['FOLIAGE_GUI_STARTED'] = 'False'      # Used by ui.py functions.

    config_signals()
    config_backup_dir(None if backup_dir == 'B' else backup_dir)
    config_credentials(None if creds_file == 'C' else creds_file, not no_keyring)
    config_port(port if (port != 'P' and isint(port)) else 8080)
    config_demo_mode(demo_mode)

    log_config()

    # Do the real work --------------------------------------------------------

    exception = exception_info = widget = None
    try:
        log('configuring PyWebIO server')
        pywebio.config(title = 'Foliage', description = 'FOLIo chAnGe Editor',
                       js_code = JS_CODE, css_style = CSS_CODE)

        # This uses a custom index page template that was created by copying
        # the PyWebIO default and modifying it.
        here = realpath(dirname(__file__))
        index_tpl = join(here, 'data', 'index.tpl')
        with open(index_tpl, encoding = 'utf-8') as index_tpl:
            log(f'reading index page template {index_tpl}')
            index_page_template = Template(index_tpl.read())
        pywebio.platform.utils._index_page_tpl = index_page_template

        # Start the widget outside the PyWebIO app so we can stop it later.
        widget = SystemWidget() if not no_widget else None

        # cdn = False makes it load PyWebIO JS code from our local copy.
        log('starting PyWebIO server')
        foliage = partial(foliage_page, widget)
        start_server(foliage, auto_open_webbrowser = True, cdn = False,
                     port = os.environ['PORT'], debug = os.environ['DEBUG'])
    except KeyboardInterrupt:
        # Catch it, but don't treat it as an error; just stop execution.
        log('keyboard interrupt received')
        pass
    except SystemExit:
        # Thrown by quit_app() during a normal exit.
        log('exit requested')
    except Exception as ex:             # noqa: PIE786
        exception = ex
        exception_info = sys.exc_info()

    # Try to deal with exceptions gracefully ----------------------------------

    if widget:
        # Stop the widget explicitly now, in case have an exception and end up
        # calling os._exit, b/c doing *that* would not kill the widget subproc.
        widget.stop()

    if exception:
        log('Main caught exception: ' + str(exception))
        from traceback import format_exception
        summary = str(exception_info[1])
        details = ''.join(format_exception(*exception_info))
        log('Exception info: ' + summary + '\n' + details)
        # Try to tell the user what happened, if we can.
        try:
            note_error('Error: ' + summary)
            wait(2)
            log('closing application window')
            run_js('close_window()')
        except Exception:               # noqa: PIE786
            pass
        log('exiting forcefully with error code')
        # This is a sledgehammer, but it kills everything, including network
        # get/post and the Qt widget thread (on Windows).
        os._exit(1)
    else:
        log('exiting normally')

    # And exit ----------------------------------------------------------------

    log('_'*8 + f' stopped {timestamp()} ' + '_'*8)


# Main page creation function.
# .............................................................................

def foliage_page(widget):
    '''Main page creation function and main loop for Foliage.
    This is handed to the PyWebIO start_server() function in our main().
    '''
    os.environ['FOLIAGE_GUI_STARTED'] = 'True'   # Used by ui.py functions.
    log('generating main Foliage page')
    put_image(image_data('foliage-icon.png'), width='70px').style('float: left')
    put_image(image_data('foliage-icon-r.png'), width='70px').style('float: right')
    put_html('<h1 class="text-center">Foliage</h1>')
    put_html('<div class="font-italic text-muted font-weight-light text-center mx-auto">'
             '<b>FOLIo chAnGe Editor</b> runs on your computer and interacts'
             ' with FOLIO over the network.').style('width: 85%')
    put_tabs([tab.contents() for tab in _TABS]).style('padding-bottom: 16px')
    put_actions('quit',
                buttons = [{'label': 'Quit', 'value': True, 'color': 'warning'}]
                ).style('position: absolute; bottom: 0px;'
                        'left: calc(50% - 3em); z-index: 2')
    warn_if_demo_mode()
    close_splash_screen()

    # Make sure we have a working network before trying to test Folio creds.
    if not network_available():
        notify('No network -- cannot proceed.')
        quit_app(ask_confirm = False)

    # Make sure we have valid FOLIO credentials.
    check_credentials()

    # Create a single dict from all the separate pin_watchers dicts.
    watchers  = dict(ChainMap(*[tab.pin_watchers() for tab in _TABS]))
    pin_names = ['quit'] + list(watchers.keys())

    log(f'entering pin handler loop for pins {pin_names}')
    while True:
        # Block, waiting for a change event on any of the pins being watched.
        # The timeout is so we can check if the user quit the taskbar widget.
        changed = pin_wait_change(pin_names, timeout = 1)
        if (not widget or widget.running()) and not changed:
            continue
        if (widget and not widget.running()):
            log('widget has exited')
            quit_app(ask_confirm = False)
        if changed and changed['name'] == 'quit':
            log('user clicked the Quit button')
            quit_app(ask_confirm = True)
            continue                    # In case the user cancels the exit.
        # Find handler associated w/ pin name & call it with value from event.
        name = changed["name"]
        log(f'invoking pin callback for {name}')
        watchers[name](changed['value'])


# Miscellaneous utilities local to this module.
# .............................................................................

def config_debug(debug_arg):
    '''Takes the value of the --debug flag & configures debugging accordingly.'''

    log_file = None
    try:
        if debug_arg == 'OUT':
            # The --debug flag was not given, so turn on only basic logging.
            # Store the debug log in user's log directory.
            log_dir = _DIRS.user_log_dir
            log_file = join(log_dir, 'log.txt')
            if not exists(log_dir):
                makedirs(log_dir)
            with open(log_file, 'w'):
                # Empty out the file for each new run.
                pass
        else:
            # We were given --debug explicitly.  Turn on all debug features.
            if debug_arg != '-':
                log_file = debug_arg
                log_dir = dirname(log_file)
                if not writable(log_dir):
                    note_error(f'Can\'t write debug ouput in {log_dir}')
                    sys.exit()
            faulthandler.enable()
            if os.name != 'nt':         # Can't use next part on Windows.
                import signal
                from boltons.debugutils import pdb_on_signal
                pdb_on_signal(signal.SIGUSR1)

            note_info('Debug & auto-reload are on. "kill -USR1 pid" for pdb.')

            # Turn on debug logging in PyWebIO & Tornado. That ends up enabling
            # logging in hpack, which is too much, so turn that off separately.
            import logging
            logging.root.setLevel(logging.DEBUG)
            logging.getLogger('hpack').setLevel(logging.INFO)

        # Turn on debug tracing to the destination we ended up deciding to use.
        set_debug(True, log_file or '-')
        log('debug_arg = ' + debug_arg)
    except PermissionError:
        note_warn(f'Permission denied creating log file {antiformat(log_file)}'
                  ' -- debug log will not be written.')
    except FileNotFoundError:
        note_warn(f'Cannot write log file {antiformat(log_file)}'
                  ' -- debug log will not be written.')
    except KeyboardInterrupt:
        # Need to catch this separately or else it will end up ignored by
        # virtue of the next clause catching all Exceptions.
        os._exit()
    except Exception as ex:             # noqa: PIE786
        log(str(ex))
        note_warn(f'Unable to create log file {antiformat(log_file)}'
                  ' -- debug log will not be written.')

    # Make settings accessible in other parts of the program.
    os.environ['DEBUG'] = str(debug_arg != 'OUT')
    os.environ['LOG_FILE'] = (log_file or '-')


def config_signals():
    '''Configure process signal handling.'''
    if os.name == 'nt' and inside_pyinstaller_app():
        # Our PyQt taskbar widget is problematic because PyQt uses signals
        # and when you exit the widget, the signal it sends causes the main
        # thread to terminate, which in turn means the main thread never gets
        # a chance to close the application window.  I couldn't solve this by
        # catching the signals, maybe because I just don't know how to do
        # that on Windows.  Instead, this causes them to be ignored.  IMHO it's
        # OK when running as a GUI app b/c it's hard for the user to ^C it.
        import signal
        import win32api
        import ctypes
        log('configuring signals to be ignored on Windows')
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        win32api.SetConsoleCtrlHandler(None, True)
        ctypes.windll.kernel32.SetConsoleCtrlHandler(None, True)
    else:
        config_interrupt(raise_exception = Interrupted)


def config_backup_dir(backup_dir):
    '''Configure the directory used for backing up FOLIO records.'''
    if not backup_dir:
        default_backups = join(_DIRS.user_data_dir, 'Backups')
        backup_dir = config('BACKUP_DIR', default = default_backups)
    if exists(backup_dir) and not isdir(backup_dir):
        note_error(f'Not a directory: {antiformat(backup_dir)}')
        sys.exit(1)
    if not exists(backup_dir):
        log(f'creating backup directory {antiformat(backup_dir)}')
        try:
            makedirs(backup_dir)
        except OSError:
            note_error(f'Unable to create backup directory {antiformat(backup_dir)}')
            sys.exit(1)
    if not writable(backup_dir):
        note_error(f'Cannot write in backup directory: {antiformat(backup_dir)}')
        sys.exit(1)
    log('backup dir is ' + backup_dir)
    os.environ['BACKUP_DIR'] = backup_dir


def config_credentials(creds_file, use_keyring):
    '''Takes credentials-related command line options and processes them.'''
    os.environ['USE_KEYRING'] = str(use_keyring)
    os.environ['CREDS_FILE'] = creds_file or 'None'

    creds = None
    if creds_file:
        log('creds file supplied on command line')
        if not exists(creds_file):
            note_error(f'Credentials file does not exist: {creds_file}')
            sys.exit(1)
        if not readable(creds_file):
            note_error(f'Credentials file not readable: {creds_file}')
            sys.exit(1)
        creds = credentials_from_file(creds_file)
        if not creds:
            note_error(f'Failed to read credentials from {creds_file}')
            sys.exit(1)
        if not credentials_complete(creds):
            # Consider it an error to be told to use a file and it's incomplete
            note_error(f'Incomplete credentials in {creds_file}')
            sys.exit(1)
    if not creds:
        log('no creds supplied on command line; looking in env')
        creds = credentials_from_env()
    keyring_creds = None
    if (not creds or not credentials_complete(creds)) and use_keyring:
        log('no creds found in env; looking in keyring')
        keyring_creds = credentials_from_keyring(partial_ok = True)
        if credentials_complete(keyring_creds):
            creds = keyring_creds
    if creds:
        log('found creds and using them')
        use_credentials(creds)
    else:
        log('no creds found')


def config_demo_mode(demo_mode):
    '''Takes the --demo-mode option and handles it.'''
    os.environ['DEMO_MODE'] = str(demo_mode)
    if demo_mode:
        note_warn('Demo mode is on -- changes to FOLIO will NOT be made')


def config_port(port):
    '''Takes the --port option and changes the Foliage port if needed.'''
    if not isint(port):
        note_error(f'Port number value is not an integer: {antiformat(port)}')
        sys.exit(1)
    os.environ['PORT'] = str(port)


def log_config():
    '''Write the configuration to the log file.'''
    import platform
    log(f'Foliage version = {__version__}')
    log(f'system          = {platform.system()}')
    if platform.system() == 'Darwin':
        log(f'version         = {platform.mac_ver()[0]}')
    else:
        log(f'version         = {platform.version()}')
    log(f'debug           = {os.environ["DEBUG"]}')
    log(f'backup_dir      = {config("BACKUP_DIR")}')
    log(f'log_file        = {config("LOG_FILE")}')
    log(f'creds_file      = {config("CREDS_FILE")}')
    log(f'use_keyring     = {config("USE_KEYRING")}')
    log(f'port            = {config("PORT")}')
    log(f'demo_mode       = {config("DEMO_MODE")}')


def warn_if_demo_mode():
    '''Put a marker on the Foliage GUI to indicate that demo mode is in effect.'''
    if config('DEMO_MODE', cast = bool):
        put_warning('Demo mode in effect').style(
            'position: absolute; left: calc(50% - 5.5em); width: 11em;'
            'height: 25px; padding: 0 10px; top: 0; z-index: 2')
    else:
        log('Demo mode not in effect')


def check_credentials():
    '''Check that the credentials we have are complete and valid.
    If they are not, ask the user if they want to edit them.
    '''
    def edit_and_use_credentials():
        creds = credentials_from_user(initial_creds = credentials_from_env())
        if not credentials_complete(creds):
            notify('Unable to proceed without complete credentials. Quitting.')
            quit_app(ask_confirm = False)
        use_credentials(creds)

    if not credentials_complete(credentials_from_env()):
        edit_and_use_credentials()
    if not Folio().credentials_valid():
        # FOLIO might have invalidated users' tokens.
        if confirm('The FOLIO token may have expired, or else the given '
                   ' credentials are invalid. Click "OK" to review the '
                   ' credentials and try to regenerate the token, or click '
                   ' "Cancel" to quit Foliage now.'):
            edit_and_use_credentials()
        else:
            notify('Invalid FOLIO credentials, or expired token.')
            quit_app(ask_confirm = False)


# Main entry point.
# .............................................................................

# The following entry point definition is for the console_scripts keyword
# option to setuptools.  The entry point for console_scripts has to be a
# function that takes zero arguments.
def console_scripts_main():
    plac.call(main)


# The following allows users to invoke this using "python3 -m handprint".
if __name__ == '__main__':
    plac.call(main)



'''
list_tab.py: implementation of the "List UUIDs" tab

Copyright
---------

Copyright (c) 2021-2022 by the California Institute of Technology.  This code
is open-source software released under a 3-clause BSD license.  Please see the
file "LICENSE" for more information.
'''

from   commonpy.exceptions import Interrupted
from   commonpy.interrupt import wait, interrupt, reset_interrupts
from   decouple import config
from   pywebio.output import put_markdown, put_row, put_button
from   pywebio.output import use_scope, clear, put_warning, put_grid
from   pywebio.output import put_processbar, set_processbar
from   pywebio.output import put_scope, clear_scope
from   pywebio.pin import pin, put_textarea
from   pywebio.session import eval_js
from   sidetrack import log

from   foliage.base_tab import FoliageTab
from   foliage.exceptions import FolioOpFailed
from   foliage.export import export_data
from   foliage.folio import Folio, RecordKind, IdKind, Record
from   foliage.folio import unique_identifiers, back_up_record
from   foliage.ui import stop_processbar, note_error, user_file
from   foliage.ui import tell_success, tell_failure, tell_warning, PROGRESS_BOX


# Tab definition class.
# .............................................................................

class CleanTab(FoliageTab):
    def contents(self):
        return {'title': 'Clean Records', 'content': tab_contents()}

    def pin_watchers(self):
        return {}


# Tab layout.
# .............................................................................

def tab_contents():
    log('generating list tab contents')
    return [
        put_markdown('### Phantom loans'),
        put_grid([[
            put_markdown("Input one or more **user barcodes** or **user id**'s"
                         ' below, or upload a file containing them. Click'
                         " the button below to delete user's loans on items"
                         " that can't be found in FOLIO. Loans on other items"
                         ' (i.e., for any items that _can_ be found in FOLIO)'
                         ' will be left untouched.'),
            put_button('Upload', outline = True,
                       onclick = lambda: load_file()).style(
                           'text-align: right'),
        ]], cell_widths = 'auto 100px'),
        put_textarea('textbox_users', rows = 4),
        put_row([
            put_button('Delete phantom loans',
                       onclick = lambda: do_delete()
                       ).style('text-align: left'),
            put_button('Clear', outline = True,
                       onclick = lambda: clear_tab()).style('text-align: right'),
        ])
    ]


# Implementation of tab functionality.
# .............................................................................

_interrupted = False
_running = False
_last_textbox = ''


def clear_tab():
    global _last_textbox
    log('clearing tab')
    clear('output')
    pin.textbox_users = ''
    _last_textbox = ''


def load_file():
    log('user requesting file upload')
    if (contents := user_file('Upload a file containing identifiers')):
        pin.textbox_users = contents


def stop():
    '''Stop an ongoing deletion by setting the _interrupted flag.'''
    global _interrupted
    global _last_textbox
    log('stopping')
    _interrupted = True
    _last_textbox = ''
    stop_processbar()
    interrupt()
    with use_scope('output'):
        tell_warning('**Stopping** ...')


def reset():
    # Reset to state where we can run new operations.
    global _interrupted
    _interrupted = False
    reset_interrupts()


def enable_delete_button(state):
    '''Enable the "delete phantom loans" button if True, disable if False.'''
    action = 'removeClass' if state else 'addClass'
    eval_js(f'''$("button:contains('Delete phantom loans')").{action}("disabled-button");''')


def wait_if_running():
    '''Check if the run state is running; if it is, wait until it changes.'''
    # The _running variable is set by do_find() when it's finished, and it's
    # set even if it gets interrupted. This is what we cue from.
    global _running
    if not _running:
        return
    enable_delete_button(False)
    stop()
    # Wait in case an ongoing delete is running, but don't wait forever.
    wait_count = 10
    while _running and wait_count > 0:
        # If the user clicks multiple times rapidly, the exception raised for
        # interrupts starts to cascade. Wrap with a try-except to avoid this.
        try:
            wait(1)
            wait_count -= 1
        except Interrupted:
            continue
    enable_delete_button(True)


_results = []


def clear_results():
    global _results
    _results = []


def record_result(record_or_id, success, notes):
    global _results
    id_ = record_or_id if isinstance(record_or_id, str) else record_or_id.id
    rec = record_or_id if isinstance(record_or_id, Record) else None
    _results.append({'id': id_, 'success': success, 'notes': notes, 'record': rec})


def succeeded(record_or_id, msg, why = ''):
    comment = (' (' + why + ')') if why else ''
    record_result(record_or_id, True, msg + comment)
    tell_success('Success: ' + msg + comment + '.')


def failed(record_or_id, msg, why = ''):
    comment = (' (' + why + ')') if why else ''
    record_result(record_or_id, False, msg + comment)
    id_ = record_or_id if isinstance(record_or_id, str) else record_or_id.id
    tell_failure(f'Failed to delete **{id_}**{comment}: ' + msg + '.')


def skipped(record_or_id, msg, why = ''):
    comment = (' (' + why + ')') if why else ''
    record_result(record_or_id, False, msg + comment)
    id_ = record_or_id if isinstance(record_or_id, str) else record_or_id.id
    tell_warning(f'Skipped **{id_}**{comment}: ' + msg + '.')


def do_delete():
    global _last_textbox
    global _results
    global _interrupted
    global _running
    log('do_delete invoked')
    wait_if_running()
    clear_results()
    reset()
    # Normally we'd want to find out if they input any identifiers, but I want
    # to detect *any* change to the input box, so this is a lower-level test.
    if not pin.textbox_users.strip():
        note_error('Please input at least one barcode or other id.')
        return
    identifiers = unique_identifiers(pin.textbox_users)
    if not identifiers:
        note_error('The input does not appear to contain identifiers or barcodes.')
        return
    _last_textbox = pin.textbox_users
    steps = len(identifiers) + 1
    folio = Folio()
    with use_scope('output', clear = True):
        put_grid([[
            put_scope('current_activity', [
                put_markdown('_This operation can take a long time. Please be patient._'
                             ).style('color: DarkOrange; margin-bottom: 0')]),
        ], [
            put_processbar('bar', init = 1/steps).style('margin-top: 11px'),
            put_button('Stop', outline = True, color = 'danger',
                       onclick = lambda: stop()).style('text-align: right'),
        ]], cell_widths = '85% 15%').style(PROGRESS_BOX)
        _running = True
        for count, user in enumerate(identifiers, start = 2):
            if _interrupted:
                break
            try:
                # Check that the kind of id we were given is really for users.
                id_kind = folio.id_kind(user)
                if id_kind is IdKind.UNKNOWN:
                    tell_failure(f'Unrecognized identifier: **{user}**.')
                    continue
                if id_kind not in [IdKind.USER_BARCODE, IdKind.USER_ID]:
                    tell_failure(f'Not a user identifier or barcode: **{user}**.')
                    continue
                deletions = []
                deletions_ids = set()
                for loan in folio.related_records(user, id_kind, RecordKind.LOAN,
                                                  open_loans_only = False):
                    if _interrupted:
                        raise Interrupted
                    item_id = loan.data['itemId']
                    item = folio.related_records(item_id, IdKind.ITEM_ID, RecordKind.ITEM)
                    if not item and item_id not in deletions_ids:
                        log(f'item {item_id} no longer exists; need delete loan {loan.id}')
                        deletions.append(loan)
                        deletions_ids.add(item_id)
                if not deletions:
                    put_warning('Did not find any loans on deleted items for'
                                f' user {user} – nothing to do.')
                    continue
                for loan in deletions:
                    if _interrupted:
                        raise Interrupted
                    delete(loan, loan.data['itemId'], user)
            except Interrupted:
                log('stopping due to interruption')
                _interrupted = True
            except Exception as ex:     # noqa: PIE786
                import traceback
                log('Exception info: ' + str(ex) + '\n' + traceback.format_exc())
                tell_failure('Error: ' + str(ex))
                stop_processbar()
                return
            finally:
                if not _interrupted:
                    set_processbar('bar', count/steps)
        stop_processbar()
        clear_scope('current_activity')
        if _interrupted:
            tell_warning('**Stopped**.')
        else:
            put_grid([[
                put_markdown('Done.'),
                put_button('Export', outline = True,
                           onclick = lambda: do_export('foliage-cleaning.csv'),
                           ).style('text-align: right')
            ]]).style('margin: 1.5em 17px auto 17px')
        _running = False


def delete(record, item_id, user_id):
    '''Low-level function to delete the given record.'''
    why = f'for loan on nonexistent item {item_id} by user {user_id}'
    if config('DEMO_MODE', cast = bool):
        log(f'demo mode in effect – pretending to delete {record.id}')
    else:
        try:
            back_up_record(record)
            folio = Folio()
            folio.delete_record(record)
        except FolioOpFailed as ex:
            failed(record, str(ex), why)
            return False
    succeeded(record, f'deleted {record.kind} record **{record.id}**', why)
    return True


def do_export(file_name):
    global _results
    # Output fields requested
    #   id
    #   success
    #   notes
    values = []
    for result in _results:
        entry = {'Loan ID'            : result['id'],
                 'Operation success'  : result['success'],
                 'Notes'              : result['notes']}
        values.append(entry)
    export_data(values, file_name)



import pytest


@pytest.fixture
def define_env_vars(monkeypatch):
    from os import path
    monkeypatch.chdir(path.dirname(__file__))

    from decouple import Config, RepositoryIni
    source = Config(RepositoryIni(source = 'settings.ini'))
    url = source.get('FOLIO_OKAPI_URL')
    tenant = source.get('FOLIO_OKAPI_TENANT_ID')
    token = source.get('FOLIO_OKAPI_TOKEN')

    monkeypatch.setenv('FOLIO_OKAPI_URL', url)
    monkeypatch.setenv('FOLIO_OKAPI_TENANT_ID', tenant)
    monkeypatch.setenv('FOLIO_OKAPI_TOKEN', token)


def test_record_id_type(define_env_vars):
    from foliage.folio import Folio, IdKind
    folio = Folio()
    assert folio.id_kind('35047019219716') == IdKind.ITEM_BARCODE
    assert folio.id_kind('d893839b-0309-4856-b496-0db89a0a6a04') == IdKind.ITEM_ID
    assert folio.id_kind('it00002135242') == IdKind.ITEM_HRID
    assert folio.id_kind('946cce1b-0451-460e-816f-51436182efaa') == IdKind.HOLDINGS_ID
    assert folio.id_kind(r'\35047018212589') == IdKind.ITEM_BARCODE
    assert folio.id_kind(r'\35047019565233') == IdKind.ITEM_BARCODE
    assert folio.id_kind('nobarcode1') == IdKind.ITEM_BARCODE
    assert folio.id_kind('nobarcode10') == IdKind.ITEM_BARCODE
    assert folio.id_kind('nobarcode10000') == IdKind.ITEM_BARCODE
    assert folio.id_kind('nobarcode10001') == IdKind.ITEM_BARCODE
    assert folio.id_kind('nobarcode10002') == IdKind.ITEM_BARCODE
    assert folio.id_kind('350470106306') == IdKind.ITEM_BARCODE
    assert folio.id_kind('350470106969') == IdKind.ITEM_BARCODE
    assert folio.id_kind('93') == IdKind.ITEM_BARCODE
    assert folio.id_kind('95') == IdKind.ITEM_BARCODE
    assert folio.id_kind('101') == IdKind.ITEM_BARCODE
    assert folio.id_kind('TEMP-D1234') == IdKind.ITEM_BARCODE
    assert folio.id_kind('TEMP-FFF123') == IdKind.ITEM_BARCODE
    assert folio.id_kind('tmp-21924070') == IdKind.ITEM_BARCODE
    assert folio.id_kind('cit.oai.caltech.folio.ebsco.com.fs00001057.17c5c348.8796.4b11.90a8.6b31ff9509ed') == IdKind.ACCESSION
    assert folio.id_kind('cit.oai.edge.caltech.folio.ebsco.com.fs00001057.17c5c348.8796.4b11.90a8.6b31ff9509ed') == IdKind.ACCESSION


def test_extracting_instance_id():
    from foliage.folio import instance_id_from_accession
    instance_id_from_accession('cit.oai.caltech.folio.ebsco.com.fs00001057.17c5c348.8796.4b11.90a8.6b31ff9509ed') == '17c5c348-8796-4b11-90a8-6b31ff9509ed'
    instance_id_from_accession('cit.oai.edge.caltech.folio.ebsco.com.fs00001057.17c5c348.8796.4b11.90a8.6b31ff9509ed') == '17c5c348-8796-4b11-90a8-6b31ff9509ed'
